@@ -22,7 +22,7 @@
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-*/
+ */
 
 #include <string.h>
 #include <errno.h>
@@ -39,6 +39,7 @@
 #include <libgnome-desktop/gnome-desktop-thumbnail.h>
 
 #include "font-model.h"
+#include "font-utils.h"
 #include "sushi-font-loader.h"
 
 struct _FontViewModelPrivate {
@@ -46,8 +47,17 @@ struct _FontViewModelPrivate {
     FcFontSet *font_list;
 
     FT_Library library;
+
+    GList *monitors;
 };
 
+enum {
+    CONFIG_CHANGED,
+    NUM_SIGNALS
+};
+
+static guint signals[NUM_SIGNALS] = { 0, };
+
 static void ensure_thumbnail (FontViewModel *self, const gchar *path);
 
 G_DEFINE_TYPE (FontViewModel, font_view_model, GTK_TYPE_LIST_STORE);
@@ -98,6 +108,8 @@ create_thumbnail (GIOSchedulerJob *job,
     }
   else
     {
+      gnome_desktop_thumbnail_factory_create_failed_thumbnail (factory,
+                                                               uri, (time_t) mtime);
       g_simple_async_result_set_op_res_gboolean (result, FALSE);
     }
 
@@ -139,6 +151,7 @@ gd_queue_thumbnail_job_for_file_finish (GAsyncResult *res)
 
 typedef struct {
     const gchar *file;
+    FT_Face face;
     GtkTreeIter iter;
     gboolean found;
 } IterForFileData;
@@ -150,15 +163,24 @@ iter_for_file_foreach (GtkTreeModel *model,
                        gpointer user_data)
 {
     IterForFileData *data = user_data;
-    gchar *font_path;
+    gchar *font_path, *font_name, *match_name;
     gboolean retval;
 
     gtk_tree_model_get (GTK_TREE_MODEL (model), iter,
+                        COLUMN_NAME, &font_name,
                         COLUMN_PATH, &font_path,
                         -1);
 
-    retval = (g_strcmp0 (font_path, data->file) == 0);
+    if (data->file) {
+        retval = (g_strcmp0 (font_path, data->file) == 0);
+    } else if (data->face) {
+        match_name = font_utils_get_font_name (data->face);
+        retval = (g_strcmp0 (font_name, match_name) == 0);
+        g_free (match_name);
+    }
+
     g_free (font_path);
+    g_free (font_name);
 
     if (retval) {
         data->iter = *iter;
@@ -168,15 +190,17 @@ iter_for_file_foreach (GtkTreeModel *model,
     return retval;
 }
 
-gboolean
-font_view_model_get_iter_for_file (FontViewModel *self,
+static gboolean
+font_view_model_get_iter_internal (FontViewModel *self,
                                    const gchar *file,
+                                   FT_Face face,
                                    GtkTreeIter *iter)
 {
     IterForFileData *data = g_slice_new0 (IterForFileData);
     gboolean found;
 
     data->file = file;
+    data->face = face;
     data->found = FALSE;
 
     gtk_tree_model_foreach (GTK_TREE_MODEL (self),
@@ -192,30 +216,88 @@ font_view_model_get_iter_for_file (FontViewModel *self,
     return found;
 }
 
+gboolean
+font_view_model_get_iter_for_file (FontViewModel *self,
+                                   const gchar *file,
+                                   GtkTreeIter *iter)
+{
+    return font_view_model_get_iter_internal
+        (self, file, NULL, iter);
+}
+
+gboolean
+font_view_model_get_iter_for_face (FontViewModel *self,
+                                   FT_Face face,
+                                   GtkTreeIter *iter)
+{
+    return font_view_model_get_iter_internal
+        (self, NULL, face, iter);
+}
+
 typedef struct {
     FontViewModel *self;
     gchar *font_path;
 } LoadThumbnailData;
 
+static void 
+load_thumbnail_data_free (LoadThumbnailData *data)
+{
+    g_object_unref (data->self);
+    g_free (data->font_path);
+    g_slice_free (LoadThumbnailData, data);
+}
+
 static void
 thumbnail_ready_cb (GObject *source,
                     GAsyncResult *res,
                     gpointer user_data)
 {
     FontViewModel *self = user_data;
-    gboolean result;
     gchar *path;
     GFile *file;
 
-    result = gd_queue_thumbnail_job_for_file_finish (G_ASYNC_RESULT (res));
+    gd_queue_thumbnail_job_for_file_finish (G_ASYNC_RESULT (res));
+    file = G_FILE (source);
+    path = g_file_get_path (file);
+    ensure_thumbnail (self, path);
 
-    if (result) {
-        file = G_FILE (source);
-        path = g_file_get_path (file);
-        ensure_thumbnail (self, path);
+    g_free (path);
+}
 
-        g_free (path);
-    }
+static void
+set_fallback_icon (FontViewModel *self,
+                   GFile *file)
+{
+    GtkTreeIter iter;
+    GtkIconTheme *icon_theme;
+    GtkIconInfo *icon_info;
+    GdkPixbuf *pix;
+    gchar *path;
+    GIcon *icon = NULL;
+
+    icon_theme = gtk_icon_theme_get_default ();
+    icon = g_content_type_get_icon ("application/x-font-ttf");
+    icon_info = gtk_icon_theme_lookup_by_gicon (icon_theme, icon, 
+                                                128, GTK_ICON_LOOKUP_GENERIC_FALLBACK);
+    g_object_unref (icon);
+
+    if (!icon_info)
+        return;
+
+    pix = gtk_icon_info_load_icon (icon_info, NULL);
+    gtk_icon_info_free (icon_info);
+
+    if (!pix)
+        return;
+
+    path = g_file_get_path (file);
+    if (font_view_model_get_iter_for_file (self, path, &iter))
+        gtk_list_store_set (GTK_LIST_STORE (self), &iter,
+                            COLUMN_ICON, pix,
+                            -1);
+
+    g_free (path);
+    g_object_unref (pix);
 }
 
 static void
@@ -226,6 +308,7 @@ pixbuf_async_ready_cb (GObject *source,
     LoadThumbnailData *data = user_data;
     GdkPixbuf *pix;
     GtkTreeIter iter;
+    GFile *file;
 
     pix = gdk_pixbuf_new_from_stream_finish (res, NULL);
 
@@ -236,11 +319,13 @@ pixbuf_async_ready_cb (GObject *source,
                                 -1);
 
         g_object_unref (pix);
+    } else {
+        file = g_file_new_for_path (data->font_path);
+        set_fallback_icon (data->self, file);
+        g_object_unref (file);
     }
 
-    g_object_unref (data->self);
-    g_free (data->font_path);
-    g_slice_free (LoadThumbnailData, data);
+    load_thumbnail_data_free (data);
 }
 
 static void
@@ -259,6 +344,9 @@ thumb_file_read_async_ready_cb (GObject *source,
                                                    128, 128, TRUE,
                                                    NULL, pixbuf_async_ready_cb, data);
         g_object_unref (is);
+    } else {
+        set_fallback_icon (data->self, G_FILE (source));
+        load_thumbnail_data_free (data);
     }
 }
 
@@ -277,10 +365,11 @@ ensure_thumbnail (FontViewModel *self,
                               G_FILE_QUERY_INFO_NONE,
                               NULL, NULL);
 
-    if (!info)
-        goto out;
-    if (g_file_info_get_attribute_boolean (info, G_FILE_ATTRIBUTE_THUMBNAILING_FAILED))
+    if (!info ||
+        g_file_info_get_attribute_boolean (info, G_FILE_ATTRIBUTE_THUMBNAILING_FAILED)) {
+        set_fallback_icon (self, file);
         goto out;
+    }
 
     thumb_path = g_file_info_get_attribute_byte_string (info, G_FILE_ATTRIBUTE_THUMBNAIL_PATH);
     if (thumb_path)
@@ -304,37 +393,6 @@ ensure_thumbnail (FontViewModel *self,
     g_clear_object (&info);
 }
 
-static gchar *
-get_font_name (FontViewModel *self,
-               const gchar *path)
-{
-    GFile *file;
-    gchar *uri, *contents = NULL, *name = NULL;
-    GError *error = NULL;
-    FT_Face face;
-
-    file = g_file_new_for_path (path);
-    uri = g_file_get_uri (file);
-
-    face = sushi_new_ft_face_from_uri (self->priv->library, uri, &contents, &error);
-    if (face != NULL) {
-        if (g_strcmp0 (face->style_name, "Regular") == 0)
-            name = g_strdup (face->family_name);
-        else
-            name = g_strconcat (face->family_name, ", ", face->style_name, NULL);
-        FT_Done_Face (face);
-    } else if (error != NULL) {
-        g_warning ("Can't get font name: %s\n", error->message);
-        g_error_free (error);
-    }
-
-    g_free (uri);
-    g_object_unref (file);
-    g_free (contents);
-
-    return name;
-}
-
 /* make sure the font list is valid */
 static void
 ensure_font_list (FontViewModel *self)
@@ -346,6 +404,17 @@ ensure_font_list (FontViewModel *self)
     FcChar8 *file;
     gchar *font_name;
 
+    if (self->priv->font_list) {
+            FcFontSetDestroy (self->priv->font_list);
+            self->priv->font_list = NULL;
+    }
+
+    gtk_list_store_clear (GTK_LIST_STORE (self));
+
+    /* always reinitialize the font database */
+    if (!FcInitReinitialize())
+        return;
+
     pat = FcPatternCreate ();
     os = FcObjectSetBuild (FC_FILE, FC_FAMILY, FC_WEIGHT, FC_SLANT, NULL);
 
@@ -359,7 +428,10 @@ ensure_font_list (FontViewModel *self)
 
     for (i = 0; i < self->priv->font_list->nfont; i++) {
 	FcPatternGetString (self->priv->font_list->fonts[i], FC_FILE, 0, &file);
-        font_name = get_font_name (self, (const gchar *) file);
+        font_name = font_utils_get_font_name_for_file (self->priv->library, (const gchar *) file);
+
+        if (!font_name)
+            continue;
 
         gtk_list_store_append (GTK_LIST_STORE (self), &iter);
         gtk_list_store_set (GTK_LIST_STORE (self), &iter,
@@ -397,6 +469,52 @@ font_view_model_sort_func (GtkTreeModel *model,
     return retval;
 }
 
+static void
+file_monitor_changed_cb (GFileMonitor *monitor,
+                         GFile *file,
+                         GFile *other_file,
+                         GFileMonitorEvent event,
+                         gpointer user_data)
+{
+    FontViewModel *self = user_data;
+
+    if (event == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT ||
+        event == G_FILE_MONITOR_EVENT_DELETED ||
+        event == G_FILE_MONITOR_EVENT_CREATED) {
+        ensure_font_list (self);
+        g_signal_emit (self, signals[CONFIG_CHANGED], 0);
+    }
+}
+
+static void
+create_file_monitors (FontViewModel *self)
+{
+    FcConfig *config;
+    FcStrList *str_list;
+    FcChar8 *path;
+    GFile *file;
+    GFileMonitor *monitor;
+
+    config = FcConfigGetCurrent ();
+    str_list = FcConfigGetFontDirs (config);
+
+    while ((path = FcStrListNext (str_list)) != NULL) {
+        file = g_file_new_for_path ((const gchar *) path);
+        monitor = g_file_monitor (file, G_FILE_MONITOR_NONE,
+                                  NULL, NULL);
+
+        if (monitor != NULL) {
+            self->priv->monitors = g_list_prepend (self->priv->monitors, monitor);
+            g_signal_connect (monitor, "changed",
+                              G_CALLBACK (file_monitor_changed_cb), self);
+        }
+
+        g_object_unref (file);
+    }
+
+    FcStrListDone (str_list);
+}
+
 static void
 font_view_model_init (FontViewModel *self)
 {
@@ -405,8 +523,6 @@ font_view_model_init (FontViewModel *self)
 
     self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, FONT_VIEW_TYPE_MODEL, FontViewModelPrivate);
 
-    if (!FcInit())
-        g_critical ("Can't initialize fontconfig library");
     if (FT_Init_FreeType (&self->priv->library) != FT_Err_Ok)
         g_critical ("Can't initialize FreeType library");
 
@@ -420,7 +536,9 @@ font_view_model_init (FontViewModel *self)
                                      COLUMN_NAME,
                                      font_view_model_sort_func,
                                      NULL, NULL);
+
     ensure_font_list (self);
+    create_file_monitors (self);
 }
 
 static void
@@ -438,6 +556,8 @@ font_view_model_finalize (GObject *obj)
         self->priv->library = NULL;
     }
 
+    g_list_free_full (self->priv->monitors, (GDestroyNotify) g_object_unref);
+
     G_OBJECT_CLASS (font_view_model_parent_class)->finalize (obj);
 }
 
@@ -447,6 +567,13 @@ font_view_model_class_init (FontViewModelClass *klass)
     GObjectClass *oclass = G_OBJECT_CLASS (klass);
     oclass->finalize = font_view_model_finalize;
 
+    signals[CONFIG_CHANGED] = 
+        g_signal_new ("config-changed",
+                      FONT_VIEW_TYPE_MODEL,
+                      G_SIGNAL_RUN_FIRST,
+                      0, NULL, NULL, NULL,
+                      G_TYPE_NONE, 0);
+
     g_type_class_add_private (klass, sizeof (FontViewModelPrivate));
 }
 
