@@ -29,6 +29,7 @@
 #include FT_SFNT_NAMES_H
 #include FT_TRUETYPE_IDS_H
 #include <cairo/cairo-ft.h>
+#include <fontconfig/fontconfig.h>
 #include <gio/gio.h>
 #include <gtk/gtk.h>
 #include <glib/gi18n.h>
@@ -71,23 +72,6 @@ G_DEFINE_TYPE (FontViewApplication, font_view_application, GTK_TYPE_APPLICATION)
 
 static void font_view_application_do_overview (FontViewApplication *self);
 
-static const gchar *app_menu =
-    "<interface>"
-    "  <menu id=\"app-menu\">"
-    "    <section>"
-    "      <item>"
-    "        <attribute name=\"action\">app.about</attribute>"
-    "	     <attribute name=\"label\" translatable=\"yes\">About Font Viewer</attribute>"
-    "      </item>"
-    "      <item>"
-    "       <attribute name=\"action\">app.quit</attribute>"
-    "	    <attribute name=\"label\" translatable=\"yes\">Quit</attribute>"
-    "      </item>"
-    "    </section>"
-    "  <menu>"
-    "</interface>";
-
-
 #define VIEW_ITEM_WIDTH 140
 #define VIEW_ITEM_WRAP_WIDTH 128
 #define VIEW_COLUMN_SPACING 36
@@ -96,7 +80,7 @@ static const gchar *app_menu =
 #define WHITESPACE_CHARS "\f \t"
 
 static void
-strip_whitespace (gchar **copyright)
+strip_whitespace (gchar **original)
 {
     GString *reassembled;
     gchar **split;
@@ -104,7 +88,7 @@ strip_whitespace (gchar **copyright)
     gint idx, n_stripped;
     size_t len;
 
-    split = g_strsplit (*copyright, "\n", -1);
+    split = g_strsplit (*original, "\n", -1);
     reassembled = g_string_new (NULL);
     n_stripped = 0;
 
@@ -115,15 +99,39 @@ strip_whitespace (gchar **copyright)
         if (len)
             str += len;
 
+        if (strlen (str) == 0 &&
+            ((split[idx + 1] == NULL) || strlen (split[idx + 1]) == 0))
+            continue;
+
         if (n_stripped++ > 0)
             g_string_append (reassembled, "\n");
         g_string_append (reassembled, str);
     }
 
     g_strfreev (split);
-    g_free (*copyright);
+    g_free (*original);
 
-    *copyright = g_string_free (reassembled, FALSE);
+    *original = g_string_free (reassembled, FALSE);
+}
+
+#define MATCH_VERSION_STR "Version"
+
+static void
+strip_version (gchar **original)
+{
+    gchar *ptr, *stripped;
+
+    ptr = g_strstr_len (*original, -1, MATCH_VERSION_STR);
+    if (!ptr)
+        return;
+
+    ptr += strlen (MATCH_VERSION_STR);
+    stripped = g_strdup (ptr);
+
+    strip_whitespace (&stripped);
+
+    g_free (*original);
+    *original = stripped;
 }
 
 static void
@@ -223,23 +231,35 @@ populate_grid (FontViewApplication *self,
 	    }
 	}
 	if (version) {
-	    add_row (grid, _("Version"), version, FALSE);
+            strip_version (&version);
+            add_row (grid, _("Version"), version, FALSE);
 	    g_free (version);
 	}
 	if (copyright) {
             strip_whitespace (&copyright);
-	    add_row (grid, _("Copyright"), copyright, TRUE);
+            add_row (grid, _("Copyright"), copyright, TRUE);
 	    g_free (copyright);
 	}
 	if (description) {
-	    add_row (grid, _("Description"), description, TRUE);
+            strip_whitespace (&description);
+            add_row (grid, _("Description"), description, TRUE);
 	    g_free (description);
 	}
     } else if (FT_Get_PS_Font_Info (face, &ps_info) == 0) {
-	if (ps_info.version && g_utf8_validate (ps_info.version, -1, NULL))
-	    add_row (grid, _("Version"), ps_info.version, FALSE);
-	if (ps_info.notice && g_utf8_validate (ps_info.notice, -1, NULL))
-	    add_row (grid, _("Copyright"), ps_info.notice, TRUE);
+        gchar *compressed;
+
+	if (ps_info.version && g_utf8_validate (ps_info.version, -1, NULL)) {
+            compressed = g_strcompress (ps_info.version);
+            strip_version (&compressed);
+            add_row (grid, _("Version"), compressed, FALSE);
+            g_free (compressed);
+        }
+	if (ps_info.notice && g_utf8_validate (ps_info.notice, -1, NULL)) {
+            compressed = g_strcompress (ps_info.notice);
+            strip_whitespace (&compressed);
+            add_row (grid, _("Copyright"), compressed, TRUE);
+            g_free (compressed);
+        }
     }
 }
 
@@ -247,23 +267,21 @@ static void
 install_button_refresh_appearance (FontViewApplication *self,
                                    GError *error)
 {
-    gchar *path;
+    FT_Face face;
 
     if (error != NULL) {
         gtk_button_set_label (GTK_BUTTON (self->install_button), _("Install Failed"));
         gtk_widget_set_sensitive (self->install_button, FALSE);
     } else {
-        path = g_file_get_path (self->font_file);
+        face = sushi_font_widget_get_ft_face (SUSHI_FONT_WIDGET (self->font_widget));
 
-        if (font_view_model_get_iter_for_file (FONT_VIEW_MODEL (self->model), path, NULL)) {
+        if (font_view_model_get_iter_for_face (FONT_VIEW_MODEL (self->model), face, NULL)) {
             gtk_button_set_label (GTK_BUTTON (self->install_button), _("Installed"));
             gtk_widget_set_sensitive (self->install_button, FALSE);
         } else {
             gtk_button_set_label (GTK_BUTTON (self->install_button), _("Install"));
             gtk_widget_set_sensitive (self->install_button, TRUE);
         }
-
-        g_free (path);
     }
 }
 
@@ -276,53 +294,104 @@ font_install_finished_cb (GObject      *source_object,
     GError *err = NULL;
 
     g_file_copy_finish (G_FILE (source_object), res, &err);
-    install_button_refresh_appearance (self, err);
 
     if (err != NULL) {
+        install_button_refresh_appearance (self, err);
+
         g_debug ("Install failed: %s", err->message);
         g_error_free (err);
     }
 }
 
+static void
+font_model_config_changed_cb (FontViewModel *model,
+                              gpointer user_data)
+{
+    FontViewApplication *self = user_data;
+
+    if (self->font_file != NULL)
+        install_button_refresh_appearance (self, NULL);
+}
+
 static void
 install_button_clicked_cb (GtkButton *button,
                            gpointer user_data)
 {
     FontViewApplication *self = user_data;
-    GFile *dest;
-    gchar *dest_path, *dest_filename;
+    gchar *dest_filename;
     GError *err = NULL;
+    FcConfig *config;
+    FcStrList *str_list;
+    FcChar8 *path;
+    GFile *xdg_prefix, *home_prefix, *file;
+    GFile *xdg_location = NULL, *home_location = NULL;
+    GFile *dest_location = NULL, *dest_file;
 
-    /* first check if ~/.fonts exists */
-    dest_path = g_build_filename (g_get_home_dir (), ".fonts", NULL);
-    if (!g_file_test (dest_path, G_FILE_TEST_EXISTS)) {
-        GFile *f = g_file_new_for_path (dest_path);
-        g_file_make_directory_with_parents (f, NULL, &err);
-        g_object_unref (f);
+    config = FcConfigGetCurrent ();
+    str_list = FcConfigGetFontDirs (config);
+
+    home_prefix = g_file_new_for_path (g_get_home_dir ());
+    xdg_prefix = g_file_new_for_path (g_get_user_data_dir ());
+
+    /* pick the XDG location, if any, or fallback to the first location
+     * under the home directory.
+     */
+    while ((path = FcStrListNext (str_list)) != NULL) {
+        file = g_file_new_for_path ((const gchar *) path);
+
+        if (g_file_has_prefix (file, xdg_prefix)) {
+            xdg_location = file;
+            break;
+        }
+
+        if ((home_location == NULL) &&
+            g_file_has_prefix (file, home_prefix)) {
+            home_location = file;
+            break;
+        }
+
+        g_object_unref (file);
+    }
+
+    FcStrListDone (str_list);
+    g_object_unref (home_prefix);
+    g_object_unref (xdg_prefix);
+
+    if (xdg_location != NULL)
+        dest_location = g_object_ref (xdg_location);
+    else if (home_location != NULL)
+        dest_location = g_object_ref (home_location);
+
+    g_clear_object (&home_location);
+    g_clear_object (&xdg_location);
+
+    if (dest_location == NULL) {
+        g_warning ("Install failed: can't find any configured user font directory.");
+        return;
+    }
+
+    if (!g_file_query_exists (dest_location, NULL)) {
+        g_file_make_directory_with_parents (dest_location, NULL, &err);
         if (err) {
             /* TODO: show error dialog */
             g_warning ("Could not create fonts directory: %s", err->message);
             g_error_free (err);
-            g_free (dest_path);
+            g_object_unref (dest_location);
             return;
         }
     }
-    g_free (dest_path);
 
     /* create destination filename */
     dest_filename = g_file_get_basename (self->font_file);
-    dest_path = g_build_filename (g_get_home_dir (), ".fonts", dest_filename,
-				  NULL);
+    dest_file = g_file_get_child (dest_location, dest_filename);
     g_free (dest_filename);
 
-    dest = g_file_new_for_path (dest_path);
-
     /* TODO: show error dialog if file exists */
-    g_file_copy_async (self->font_file, dest, G_FILE_COPY_NONE, 0, NULL, NULL, NULL,
+    g_file_copy_async (self->font_file, dest_file, G_FILE_COPY_NONE, 0, NULL, NULL, NULL,
                        font_install_finished_cb, self);
 
-    g_object_unref (dest);
-    g_free (dest_path);
+    g_object_unref (dest_file);
+    g_object_unref (dest_location);
 }
 
 static void
@@ -345,6 +414,7 @@ font_widget_loaded_cb (SushiFontWidget *font_widget,
 
     gd_main_toolbar_set_labels (GD_MAIN_TOOLBAR (self->toolbar),
                                 face->family_name, face->style_name);
+    install_button_refresh_appearance (self, NULL);
 }
 
 static void
@@ -385,13 +455,6 @@ static void
 font_view_application_do_open (FontViewApplication *self)
 {
     gchar *uri;
-    GtkWidget *font_widget;
-    GdkColor white = { 0, 0xffff, 0xffff, 0xffff };
-    GtkWidget *w;
-
-    w = gtk_bin_get_child (GTK_BIN (self->swin_preview));
-    if (w)
-        gtk_widget_destroy (w);
 
     self->info_button = gd_main_toolbar_add_button (GD_MAIN_TOOLBAR (self->toolbar),
                                                     NULL, _("Info"), 
@@ -403,7 +466,6 @@ font_view_application_do_open (FontViewApplication *self)
     self->install_button = gd_main_toolbar_add_button (GD_MAIN_TOOLBAR (self->toolbar),
                                                        NULL, _("Install"), 
                                                        FALSE);
-    install_button_refresh_appearance (self, NULL);
     g_signal_connect (self->install_button, "clicked",
                       G_CALLBACK (install_button_clicked_cb), self);
 
@@ -416,20 +478,31 @@ font_view_application_do_open (FontViewApplication *self)
     gtk_widget_set_vexpand (self->toolbar, FALSE);
 
     uri = g_file_get_uri (self->font_file);
-    self->font_widget = font_widget = GTK_WIDGET (sushi_font_widget_new (uri));
 
-    gtk_widget_modify_bg (font_widget, GTK_STATE_NORMAL, &white);
+    if (self->font_widget == NULL) {
+        GdkRGBA white = { 1.0, 1.0, 1.0, 1.0 };
+        GdkRGBA black = { 0.0, 0.0, 0.0, 1.0 };
+        GtkWidget *w;
+
+        self->font_widget = GTK_WIDGET (sushi_font_widget_new (uri));
+
+        gtk_widget_override_color (self->font_widget, GTK_STATE_NORMAL, &black);
+        gtk_widget_override_background_color (self->font_widget, GTK_STATE_FLAG_NORMAL, &white);
+
+        w = gtk_viewport_new (NULL, NULL);
+        gtk_viewport_set_shadow_type (GTK_VIEWPORT (w), GTK_SHADOW_NONE);
+
+        gtk_container_add (GTK_CONTAINER (w), self->font_widget);
+        gtk_container_add (GTK_CONTAINER (self->swin_preview), w);
+
+        g_signal_connect (self->font_widget, "loaded",
+                          G_CALLBACK (font_widget_loaded_cb), self);
+    } else {
+        g_object_set (self->font_widget, "uri", uri, NULL);
+    }
+
     g_free (uri);
 
-    w = gtk_viewport_new (NULL, NULL);
-    gtk_viewport_set_shadow_type (GTK_VIEWPORT (w), GTK_SHADOW_NONE);
-
-    gtk_container_add (GTK_CONTAINER (w), font_widget);
-    gtk_container_add (GTK_CONTAINER (self->swin_preview), w);
-
-    g_signal_connect (font_widget, "loaded",
-                      G_CALLBACK (font_widget_loaded_cb), self);
-
     gtk_widget_show_all (self->swin_preview);
     gtk_notebook_set_current_page (GTK_NOTEBOOK (self->notebook), 1);
 }
@@ -533,6 +606,20 @@ font_view_application_do_overview (FontViewApplication *self)
     gtk_widget_show_all (self->swin_view);
 }
 
+static gboolean
+font_view_window_key_press_event_cb (GtkWidget *widget,
+                                     GdkEventKey *event,
+                                     gpointer user_data)
+{
+    if (event->keyval == GDK_KEY_q &&
+        (event->state & GDK_CONTROL_MASK) != 0) {
+        gtk_widget_destroy (widget);
+        return TRUE;
+    }
+
+    return FALSE;
+}
+
 static void
 font_view_application_open (GApplication *application,
                             GFile **files,
@@ -597,7 +684,7 @@ font_view_application_startup (GApplication *application)
     g_action_map_add_action_entries (G_ACTION_MAP (self), action_entries, 
                                      G_N_ELEMENTS (action_entries), self);
     builder = gtk_builder_new ();
-    gtk_builder_add_from_string (builder, app_menu, -1, NULL);
+    gtk_builder_add_from_resource (builder, "/org/gnome/font-viewer/font-view-app-menu.ui", NULL);
     menu = G_MENU_MODEL (gtk_builder_get_object (builder, "app-menu"));
     gtk_application_set_app_menu (GTK_APPLICATION (self), menu);
 
@@ -605,6 +692,8 @@ font_view_application_startup (GApplication *application)
     g_object_unref (menu);
 
     self->model = font_view_model_new ();
+    g_signal_connect (self->model, "config-changed",
+                      G_CALLBACK (font_model_config_changed_cb), self);
 
     self->main_window = window = gtk_application_window_new (GTK_APPLICATION (application));
     gtk_window_set_resizable (GTK_WINDOW (window), TRUE);
@@ -612,6 +701,9 @@ font_view_application_startup (GApplication *application)
     gtk_window_set_hide_titlebar_when_maximized (GTK_WINDOW (window), TRUE);
     gtk_window_set_title (GTK_WINDOW (window), _("Font Viewer"));
 
+    g_signal_connect (window, "key-press-event",
+                      G_CALLBACK (font_view_window_key_press_event_cb), self);
+
     self->main_grid = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
     gtk_container_add (GTK_CONTAINER (self->main_window), self->main_grid);
 
@@ -649,6 +741,14 @@ font_view_application_activate (GApplication *application)
     font_view_application_do_overview (self);
 }
 
+static void
+font_view_application_quit_mainloop (GApplication *application)
+{
+    G_APPLICATION_CLASS (font_view_application_parent_class)->quit_mainloop (application);
+
+    FcFini ();
+}
+
 static void
 font_view_application_dispose (GObject *obj)
 {
@@ -675,6 +775,7 @@ font_view_application_class_init (FontViewApplicationClass *klass)
     aclass->activate = font_view_application_activate;
     aclass->startup = font_view_application_startup;
     aclass->open = font_view_application_open;
+    aclass->quit_mainloop = font_view_application_quit_mainloop;
 
     oclass->dispose = font_view_application_dispose;
 }
@@ -683,6 +784,10 @@ static GApplication *
 font_view_application_new (void)
 {
     g_type_init ();
+
+    if (!FcInit ())
+        g_critical ("Can't initialize fontconfig library");
+
     return g_object_new (FONT_VIEW_TYPE_APPLICATION,
                          "application-id", "org.gnome.font-viewer",
                          "flags", G_APPLICATION_HANDLES_OPEN,
